# -*- coding: utf-8; -*-
#
# (c) 2004-2007 Linbox / Free&ALter Soft, http://linbox.com
# (c) 2007-2010 Mandriva, http://www.mandriva.com
#
# $Id$
#
# This file is part of Mandriva Management Console (MMC).
#
# MMC is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# MMC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with MMC.  If not, see <http://www.gnu.org/licenses/>.

"""
MDS proxy plugin for the MMC agent.
"""

import os
import logging
import socket
import fcntl
import struct
import re
from mmc.core.version import scmRevision
from mmc.support import mmctools
from mmc.support.config import PluginConfig, ConfigException
from mmc.plugins.base import LogView
from ConfigParser import NoSectionError, NoOptionError

from mmc.core.audit import AuditFactory as AF
from mmc.plugins.proxy.audit import AT, AA, PLUGIN_NAME

from mmc.support.mmctools import ServiceManager, shlaunch, shlaunchDeferred

VERSION = "2.5.3"
APIVERSION = "1:1:0"
REVISION = scmRevision("$Rev$")

def getVersion(): return VERSION
def getApiVersion(): return APIVERSION
def getRevision(): return REVISION

def activate():
    """
     this function define if the module "proxy" can be activated.
     @return: return True if this module can be activate
     @rtype: boolean
    """
    config = ProxyConfig("proxy")
    logger = logging.getLogger()
    if config.disabled:
        msg = "disabled by configuration"        
        logger.warning("Plugin proxy: " + msg + ".")
        return False
    result = True
    msg = ""
    try:
        config.check()
    except ConfigException, ce:
        msg = str(ce)
        result = False
    except Exception, e:
        msg = str(e)
        result = False
    if not result:
        logger.warning("Plugin proxy: " + msg + ".")
    return result

def getBlackList():
    return Blacklist().getBlacklist()

def addBlackList(elt):
    return Blacklist().addBlacklist(elt)

def delBlackList(elt):
    return Blacklist().delBlacklist(elt)

def getStatutProxy():
    return Blacklist().getStatutProxy()

def restartSquid():
    return Blacklist().restartSquid()

def squidService(command):
#    ret = ''
#    if command == 'restart':
#      r = AF().log(PLUGIN_NAME, AA.PROXY_RESTART_SQUID_SERVICE,)
#      mmctools.shlaunchBackground(ProxyConfig("proxy").squidInit+' restart')
#    else:
    if command != 'status':
        event = { 'start' : AA.PROXY_START_SQUID_SERVICE,
                  'stop' : AA.PROXY_STOP_SQUID_SERVICE,
                  'restart' : AA.PROXY_RESTART_SQUID_SERVICE,
                  'reload' : AA.PROXY_RELOAD_SQUID_SERVICE }
    r = AF().log(PLUGIN_NAME, event[command])
    ret = SquidService().command(command)
    ret = mmctools.shlaunchBackground(ProxyConfig("proxy").squidInit+' ' +command)
    if command != 'status':
      r.commit()
    return ret

def getNetworkInterfaces():
    return NetworkManagement().getNetworkInterfaces()

def getNetworkAddr(interface = ''):
    return NetworkManagement().getIpAddr(interface)

def getSquidMode():
    return SquidManager().mode

def getSquidPort():
    return SquidManager().port

def getSquidIp():
    return SquidManager().addr

def transparentModeOn(eth_int, eth_loc, port):
    r = AF().log(PLUGIN_NAME, AA.PROXY_TRANSPARENT_MODE_ON)
    network = NetworkManagement()
    ip_loc = network.getIpAddr(eth_loc)
    mode = "http_port " + ip_loc + ":" + port + " intercept"
    SquidManager().setTransparentMode(mode)
    network.changeForwarding('1')
    network.iptablesClear()
    network.iptablesAddNat(eth_int, eth_loc, port)
    network.iptablesSave()
    squidService('reload')
    r.commit()

def transparentModeOff(port):
    r = AF().log(PLUGIN_NAME, AA.PROXY_TRANSPARENT_MODE_OFF)
    network = NetworkManagement()
    mode = "http_port " + port
    SquidManager().setTransparentMode(mode)
    network.iptablesClear()
    network.iptablesSave()
    network.changeForwarding('0')
    squidService('reload')
    r.commit()

def getSquidLog(filter = ''):
    return SquidLogView().getLog(filter)


####################################################################
#           Service management class for squid
####################################################################
class SquidService(ServiceManager):

    def __init__(self, conffile = None):
        self.config = ProxyConfig("proxy")
        ServiceManager.__init__(self, self.config.squidPidFile, self.config.squidInit)


####################################################################
#           Logger class for squid
####################################################################
class SquidLogView(LogView):
    """
    Get DHCP service log content.
    """
    def __init__(self):
        self.config = ProxyConfig("proxy")
        self.logfile = self.config.squidLogFile
        self.maxElt= 200
        self.file = open(self.logfile, 'r')
        self.pattern = {
#            "squid-syslog" : "^(?P<b>[A-z]{3}) *(?P<d>[0-9]+) (?P<H>[0-9]{2}):(?P<M>[0-9]{2}):(?P<S>[0-9]{2}) .* squid: (?P<extra>.*)$",
            "squid-syslog" : "^(?P<b>[A-z]{3}) *(?P<d>[0-9]+) (?P<H>[0-9]{2}):(?P<M>[0-9]{2}):(?P<S>[0-9]{2}) .* squid\[[0-9]+\]: (?P<extra>.*)$",
        }


####################################################################
#           Parser .ini file class for squid
####################################################################
class ProxyConfig(PluginConfig):

    def readConf(self):
        PluginConfig.readConf(self)
        try: self.sgBinary = self.get("squidguard", "path")
        except (NoSectionError, NoOptionError): pass
        try: self.sgBlacklist = self.get("squidguard", "blacklist")
        except (NoSectionError, NoOptionError): pass
        try: self.squidReload = self.get("squidguard", "scriptReload")
        except (NoSectionError, NoOptionError): pass
        try: self.squidUser = self.get("squidguard", "user")
        except (NoSectionError, NoOptionError): pass
        try: self.squidGroup = self.get("squidguard", "group")
        except (NoSectionError, NoOptionError): pass
        try: self.squidPidFile = self.get("squid", "pidfile")
        except (NoSectionError, NoOptionError): pass
        try: self.squidInit = self.get("squid", "init")
        except (NoSectionError, NoOptionError): pass
        try: self.squidLogFile = self.get("squid", "logfile")
        except (NoSectionError, NoOptionError): pass
        try: self.squidConfig = self.get("squid", "config")
        except (NoSectionError, NoOptionError): pass


    def setDefault(self):
        PluginConfig.setDefault(self)
        self.squidReload = "/etc/init.d/squid reload"
        self.squidUser = "proxy"
        self.squidGroup = "proxy"
        self.sgBinary = "/usr/bin/squidGuard"
        self.sgBlacklist = "/var/lib/squidguard/db/bad.destdomainlist"

    def check(self):
        if not os.path.exists(self.sgBinary):
            raise ConfigException("Can't find squidguard binary: " + self.sgBinary)
        # Try to get squidguard version string
        cmd = mmctools.shLaunch(self.sgBinary + " -v")
        if cmd.exitCode:
            raise ConfigException("Can't start '" + self.sgBinary + " -v': " + cmd.err + "(" + str(cmd.exitCode) + ")")
        self.sgVersion = cmd.err.strip()
        if not os.path.exists(self.sgBlacklist):
            raise ConfigException("Can't find squidguard blacklist: " + self.sgBlacklist)


####################################################################
#           Blacklist class for squid
####################################################################

class Blacklist:

    PLACEHOLDER = "placeholder"

    def __init__(self):
        """
        Load bad.destdomainlist in self.contentArr.
        For easier modification bad.destdomainlist is always loaded
        in self.contentArr.
        """
        self.config = ProxyConfig("proxy")
        self.blacklist = self.config.sgBlacklist
        self.squidguard = self.config.sgBinary
        self.contentArr = []
        f = open(self.blacklist)
        for line in f:
            line = line.strip()
            if line and line not in self.contentArr:
                self.contentArr.append(line)
        f.close()
        if not self.PLACEHOLDER in self.contentArr:
            self.contentArr.append(self.PLACEHOLDER)

    def restartSquid(self):
        """
        Compile the blacklist database and reload squid.
        """
        r = AF().log(PLUGIN_NAME, AA.PROXY_RESTART_SQUID)
        mmctools.shlaunch(self.config.sgBinary + ' -C ' + self.config.sgBlacklist)
        mmctools.shlaunch("chown " + self.config.squidGroup + "." + self.config.squidUser + " " + self.config.sgBlacklist + "*")

        psout = os.popen(self.config.squidReload, 'r')
        read = psout.read()

        if psout.close(): read = "error reloading squid"
        r.commit()

        return read

    def getStatutProxy(self):
        """
        Return proxy status
        Return if process squid and squidGuard are launched
        """
        # FIXME: works well but ugly
        res={}
        res['squid']=0
        #res['squidGuard']=0

        psout = os.popen('ps ax | grep squid | grep -v grep','r')
        try:
            tmp=psout.read()
        except:
            return res

        for a in tmp.split("\n"):
            if 'squid' in a : res['squid'] = 1
        #    if 'squidGuard' in a : res['squidGuard'] = 1
        psout.close()
        return res

    def getBlacklist(self):
        """
        Return blacklist content, without the placeholder element which must
        not be seen.
        """
        ret = self.contentArr[:]
        ret.remove(self.PLACEHOLDER)
        return ret

    def saveBlacklist(self):
        """save self.contentArr in bad.destdomainlist"""
        f = open(self.config.sgBlacklist, 'w')
        for elt in self.contentArr:
            f.write(elt + '\n')
        f.close()
        return 0

    def addBlacklist(self, elt):
        """Add an element to the blacklist"""
        r = AF().log(PLUGIN_NAME, AA.PROXY_ADD_BLACKLIST, [(elt, AT.BLACKLIST)])
        if not elt in self.contentArr:
            self.contentArr.append(elt)
        self.saveBlacklist()
        r.commit()

    def delBlacklist(self, elt):
        """Remove an element from the blacklist"""
        r = AF().log(PLUGIN_NAME, AA.PROXY_DEL_BLACKLIST, [(elt, AT.BLACKLIST)])
        if elt in self.contentArr:
            self.contentArr.remove(elt)
        self.saveBlacklist()
        r.commit()

####################################################################
#           Class for network management
####################################################################
class NetworkManagement():

    def __init__(self):
        self.iptables_save = "service iptables save"
        self.iptables_clear_nat = "iptables -t nat -F"

    def iptablesSave(self):
        return shlaunch(self.iptables_save)

    def iptablesClear(self):
        return shlaunch(self.iptables_clear_nat)

    def iptablesAddNat(self, eth_int, eth_loc, port):
        dest = self.getIpAddr(eth_loc) + ":"+ port
        ip_int = self.getIpAddr(eth_int)
        cmd = "iptables -t nat -A PREROUTING -i " + eth_int + " -p tcp ! -d " + ip_int + " --dport 80 -j DNAT --to " + dest
        shlaunch(cmd)
        cmd =  "iptables -t nat -A PREROUTING -i " + eth_loc + " -p tcp --dport 80 -j REDIRECT --to-port " + port
        shlaunch(cmd)

    def getNetworkInterfaces(self):
        """Return all network interfaces"""
        in_list = []
        f = open('/proc/net/dev', 'r')
        while True:
           line = f.readline().strip()
           if line == '':
              break
           pos = line.find(':')
           if pos == -1:
              continue
           else:
              in_list.append(line[0:pos])
        return in_list
        f.close()

    def getIpAddr(self, if_name):
        """ Return ip-address of interface. if-name - interface name """
        try:
           s = None
           s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
           if_name_byte = struct.pack('256s', if_name.encode('utf-8'))
           if_addr_byte = fcntl.ioctl(s.fileno(), 0x8915, if_name_byte)[20:24]
           return socket.inet_ntoa(if_addr_byte)
        except IOError as error:
           print('Error: {0}'.format(error))
           return None
        finally:
           if s is not None:
             s.close()

    def changeForwarding(self, val):
        """ On/off ip forward. Param val=0 forvard is off, 1 - on """
        f = open('/etc/sysctl.conf','r')
        it=0
        lines = f.readlines()
        for line in lines:
           line = line.strip()
           if re.search('net.ipv4.ip_forward',line):
             line = line[:-1] + val
             lines[it] = line + "\n"
           it += 1
        f.close()
        f = open('/etc/sysctl.conf','w')
        f.writelines(lines)
        f.close()
        shlaunch("sysctl -p")

####################################################################
#           Class work with squid.conf file
####################################################################
class SquidManager():

    def __init__(self):
        self.config = ProxyConfig("proxy").squidConfig
        self.addr = ""
        f = open(self.config)
        for line in f:
            line = line.strip()
            if re.search('http_port',line):
               if re.search("intercept", line):
                  self.mode = "on"
                  for part in  line.split(" "):
                     pos = part.find(':')
                     if pos == -1:
                       continue
                     else:
                       self.addr = part[0:pos]
                       self.port = part[pos+1:]
               else:
                  self.mode = "off"
                  pos = line.find(' ')
                  if pos != -1:
                     self.port = line[pos+1:]
        f.close()

    def setTransparentMode(self, mode):
        f = open(self.config,'r')
        it = 0
        lines = f.readlines()
        for line in lines:
           line = line.strip()
           if re.search('http_port',line):
              lines[it] = mode + "\n"
           it += 1
        f.close()
        f = open(self.config,'w')
        f.writelines(lines)
        f.close()


